import dill

from abc import ABCMeta, abstractmethod


class DataObfuscator(metaclass=ABCMeta):
    """Interface for data hashers

    Obfuscation is the method to make easily comparable, storage efficent, potentially secure versions of the data.
    """
    @staticmethod
    def _hashable_representation(data):
        """Generate a hashable representation of the data. Current implentation is pickle based.
        """
        # in the future, a "library" of representation generators could be usable
        # for efficiency gains
        return dill.dumps(data)

    @abstractmethod
    def obfuscate(self, data):
        """Create the hashed data representation

        This is method must be overridden by subclasses.

        :param data: data to obfuscate

        :return: obfuscated representation such as a hash
        """
        pass


from abc import ABCMeta, abstractmethod


class ParameterStore(metaclass=ABCMeta):
    """Interface of a parameter store object.  ParameterStores are responsible for persisting parameter values
    and determining parameter equality.
    """

    @abstractmethod
    def parameter_keys(self, key):
        """Get the parameter sub-keys for the given generator key

        Must be overridden by subclasses.

        :param key: Generator key
        :return: list of key
        """
        pass

    @abstractmethod
    def compare(self, value, parameter_key):
        """Compare the stored value to the given value

        Must be overridden by subclasses.

        :param value: value comparable to stored value
        :param parameter_key: key of value to compare to
        :return: equality test
        """
        pass

    @abstractmethod
    def update(self, key, value, parameter_keys=None):
        """ Store new value for the paramater.

        Must be overridden by subclasses.

        :param key: key  of object to store
        :param value: value of object to store
        :param parameter_keys: keys of parameters for generator parameter values
   
        """
        pass

    @abstractmethod
    def __contains__(self, key):
        """Needed for testing via in operator

        :param key:
        :return: if parameter is in this store
        """
        pass


from parsimony.generators import Generator
import pickle
import os
import string


class PickledCallableWrapper(Generator):
    """ Generator for callables using pickle as the underlying storage mechanism

    """
    def __init__(self, directory, key, function, **parameters):
        """The PickledCallableWrapper is a simple way to cache arbitrary function results.

        :param directory: directory to cache in
        :param key: generator key string
        :param function: callable handle
        :param parameters: key-value parameters for the callable function
        """
        self._function = function  # keep separate references to avoid copying later
        self._param_keys = list(parameters.keys())
        if not os.path.exists(directory):
            os.makedirs(directory)
        self._store_location = self._generate_store_location(directory, key)
        super(PickledCallableWrapper, self).__init__(key, function=function, **parameters)

    def up_to_date(self):
        """ Returns true since the framework will always take care of updates to arguments or the callable.

        :return: True
        """
        return True

    def rebuild(self):
        """ Call the callable with parameters to generate the value

        :return: generated value
        """
        params = {key: self.get_parameter(key) for key in self._param_keys}
        return self._function(**params)

    def load(self):
        """Retrieves the value from the pickle store

        :return: loaded generated value
        """
        with open(self._store_location, 'rb') as result_file:
            result = pickle.load(result_file)
        return result

    def store(self, value):
        """Dump the value to the pickle store

        :param value: value to dump
        """
        with open(self._store_location, 'wb') as result_file:
            pickle.dump(value, result_file)


    # noinspection PyMethodMayBeStatic
    def _generate_store_location(self, directory, key):
        """Generates a location in the pickle directory to store to

        :param directory: base directory
        :param key: value key, necessary because this is called in the constructor
        """
        valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
        file_name = ''.join(c if c in valid_chars else '_' for c in key)
        return os.path.join(directory, file_name)


"""
This module contains routines to fetch the default or configured implementations of
various parsimony base objects. Currently, these objects are all singletons and lazily initialized.
"""
import parsimony
import os

__store = None
# TODO, move any directory creation or other initialization into the configurable objects


def store():
    """Gets the configured parameter store. Currently, a default of PickledParameterStore is chosen.

    :return store: ParameterStore object
    """
    global __store
    if not os.path.exists('.parsimony'):
        os.makedirs('.parsimony')
    if __store is None:
        __store = parsimony.persistence.PickledParameterStore('.parsimony/p_store')
    return __store


__obfuscator = None


def obfuscator():
    """Gets the configured obfuscator. Currently, a default of SHA512Obfuscator is chosen.

    :return obfuscator: Obfuscator object
    """

    global __obfuscator
    if __obfuscator is None:
        __obfuscator = parsimony.persistence.SHA512Obfuscator()
    return __obfuscator


def callable_wrapper(key, function, **parameters):
    """Gets the configured call wrapper. Currently, a default of PickledCallableWrapper is chosen.

    :return callable_wrapper: CallableWrapper Generator object
    """
    if not os.path.exists('.parsimony/wrapped_results'):
        os.makedirs('.parsimony/wrapped_results')

    return parsimony.generators.PickledCallableWrapper('.parsimony/wrapped_results', key, function, **parameters)


from parsimony.generators import Generator
from parsimony import ParsimonyException

import os.path


class PathMonitor(Generator):
    """Monitors the path for changes. This is primarily useful as a parameter key

    """
    def __init__(self, key, file_path):
        """

        :param key: generator key
        :param file_path: file to track and generate
        """
        self._modtime = None
        super(PathMonitor, self).__init__(key, file_path=file_path)

    def up_to_date(self):
        """Determines if the file has been modified.

        :return: If file timestamp has changed or not
        """
        new_mod_time = os.path.getmtime(self.get_parameter('file_path'))
        return new_mod_time == self._modtime

    def store(self, value):
        """This is by definition already stored. Do nothing.
        """
        pass

    def rebuild(self):
        """Loads the file data.

        :return: self.load()
        """
        return self.load()

    def load(self):
        """Reads all of the file contents into a string.

        :return: text file contents
        """
        self._modtime = os.path.getmtime(self.get_parameter('file_path'))

        return self.get_parameter('file_path')


class TextFile(Generator):
    """Allows a text file to be treated as a generateable object.

    The mechanism is to load the file once and any time the file has been subsequently modified.
    """

    def __init__(self, key, file_path):
        """

        :param key: generator key
        :param file_path: file to track and generate
        """
        self._file_path_monitor = PathMonitor('text_file_path_monitor', file_path)
        super(TextFile, self).__init__(key, file_path=self._file_path_monitor)

    def up_to_date(self):
        """Determines if the file has been modified.

        :return: If file timestamp has changed or not
        """
        return self._file_path_monitor.up_to_date()

    def rebuild(self):
        """Loads the file data.

        :return: self.load()
        """
        return self.load()

    def load(self):
        """Reads all of the file contents into a string.

        :return: text file contents
        """
        with open(self._file_path_monitor.load(), 'r') as file_handle:
            contents = file_handle.read()

        return contents

    def store(self, value):
        """This is by definition already stored. Do nothing.
        """
        pass







import hashlib
from .data_obfuscator import DataObfuscator


class SHA512Obfuscator(DataObfuscator):
    """DataObfuscator that uses the SHA512 hashing mechanism.

    """

    def __init__(self):
        """
        """
        super(SHA512Obfuscator, self).__init__()

    def obfuscate(self, data):
        """Generate SHA512 of the hashable data representation

        :param data: data to obfuscate
        :return: hash
        """
        hashable_representation = self._hashable_representation(data)
        hasher = hashlib.sha512()
        hasher.update(hashable_representation)
        return hasher.digest()


import pickle
from . import ParameterStore


class PickledParameterStore(ParameterStore):
    """ParameterStore that stores values in pickled format.
    """

    def __init__(self, file_name):
        """
        Constructor
        """
        self._file_name = file_name
        try:
            with open(self._file_name, 'rb') as pickle_file:
                self._store_data = pickle.load(pickle_file)
        except IOError:  # the file didn't exist or is otherwise inaccessible, we have no choice to regenerate
            self._store_data = {}
        super(PickledParameterStore, self).__init__()

    def __contains__(self, key):
        return key in self._store_data

    def parameter_keys(self, key):
        if key in list(self._store_data.keys()):
            return self._store_data[key]['parameters']
        return {}

    def compare(self, value, parameter_key):
        return value == self._store_data[parameter_key]['value']

    def update(self, key, value, parameter_keys=None):
        self._store_data[key] = {'parameters': parameter_keys, 'value': value}
        with open(self._file_name, 'wb') as pickle_file:
            pickle.dump(self._store_data, pickle_file,protocol=pickle.HIGHEST_PROTOCOL)



from setuptools import setup, find_packages
import parsimony
setup(
    name="parsimony",
    version=parsimony.__version__,
    packages=find_packages(),
    install_requires=['dill>=0.2.1'],
    description="A library for simple caching and lazy initialization."
)


"""
Created on Jun 30, 2013

@author: rfeather
"""
import parsimony
from abc import ABCMeta, abstractmethod

GENERATOR_DEFAULT_STORE_VALUE = True


class Generator(metaclass=ABCMeta):
    """Generator abstract base class. This is the core object for cached generation.

    """

    def __init__(self, key, **parameters):
        """Generate an object.

        A generator performs parameter sensitive caching.

        :param key: string key to use when referencing the cached object
        :param parameters: parameters used to monitor the object. Parameters can be other generators, in which case recursive
        checks and generation occur as necessary.
        """
        self._key = key
        self._current_parameters = parameters
        self._generated_value = None
        self._generated = False
        self._store_keys = self._generate_store_keys()

    def generate(self):
        """Generate the object referred to by the key.

        If the object has never been generated before, the necessary work to create it is performed and the object is
        cached in a ParameterStore.

        If the object has been generated and the parameters have not been changed, the Generator checks for an in memory
         version of the object first and returns it.  If not in memory, the generated object is retrieved from a
         ParameterStore and cached in memory, then returned.

        If the object has been generated, but the parameters have changed, the object is regenerated and the ParameterStore
        is overwritten.

        :return generated_value: The value for this generator's key.
        """
        # regenerate if#
        # parameters have changed

        # did not exist in keys
        # generator parameters needed updated
        parameters_store = parsimony.configuration.store()
        obfuscator = parsimony.configuration.obfuscator()
        
        if (self._key in parameters_store) and self._parameters_up_to_date(parameters_store,obfuscator) and self.up_to_date():
            if self._generated:
                return self._generated_value
            else:
                self._generated_value = self.load()
                self._generated = True
                return self._generated_value
        else:
            for parameter, value in self._current_parameters.items():
                if not isinstance(value, parsimony.generators.Generator):
                    parameters_store.update(self._store_keys[parameter], obfuscator.obfuscate(value))
            self._update_parameters()  # parameters update needs to happen after storage since Generators get replaced

            self._generated_value = self.rebuild()
            
            parameters_store.update(self._key, obfuscator.obfuscate(GENERATOR_DEFAULT_STORE_VALUE), list(self._store_keys.values()))
            self.store(self._generated_value)
            self._generated = True
            return self._generated_value

    def _mangled_parameter_key(self, parameter_key):
        """
        Create the mapping key from this generator to sub keys
        :param parameter_key:
        :return: combined self key and parameter key
        """
        return self.key() + ':' + parameter_key

    def _generate_store_keys(self):
        """Get store keys from this objects current parameters.

        If the parameter is a generator, the key() method is used. Otherwise, the parameter key is generated in mangled
        fashion.
        :return: dict of parameter_key to key mappings
        """
        store_keys = {}
        for parameter_key, parameter in self._current_parameters.items():
            if isinstance(parameter, parsimony.generators.Generator):
                store_keys[parameter_key] = parameter.key()
            else:
                store_keys[parameter_key] = self._mangled_parameter_key(parameter_key)

        return store_keys

    @abstractmethod
    def up_to_date(self):
        """Indicate if the generated value is fresh or stale with respect to non-parmeter changes. In the case
        of completely parameter driven values this will always be true. In the case of external objects or persisted
        data, it must be checked appropriately.

        This must be overridden in subclasses.
        :return: True if the value is fresh.
        """
        pass

    @abstractmethod
    def rebuild(self):
        """Internal method to regenerate the value based on current parameters.

        This method must be overriden in subclasses. Do not use externally.

        :return: generated value
        """
        pass

    @abstractmethod
    def load(self):
        """Internal method to load the value from storage.

        This method must be overriden in subclasses. Do not use externally.

        :return: generated value
        """
        pass

    @abstractmethod
    def store(self, value):
        """Internal method to store the value in storage.

        This method must be overriden in subclasses. Do not use externally.

        :return: generated value
        """
        pass

    def _store_matches_current(self, parameters_store,obfuscator):
        """Compares the parameter_store values to the current parameters.

        Checks store key equality and value equality of non-Generator parameters. Generator parameters
        are handled separately.
        :param parameters_store:
        :return: exact match or not of parameter store key-values
        """
        # same keys
        store_parameters_keys = parameters_store.parameter_keys(self._key)
        if store_parameters_keys is None:
            return self._store_keys.values is None
        store_keys = set(store_parameters_keys)
        current_keys = set(self._store_keys.values())
        if store_keys != current_keys:
            return False

        all_values_equal = True
        # compare values and types
        for parameter_key, value in self._current_parameters.items():
            if not isinstance(value, parsimony.generators.Generator):
                all_values_equal &= parameters_store.compare(obfuscator.obfuscate(value), self._store_keys[parameter_key])

        return all_values_equal

    def key(self):
        """Get the key of this Generator

        :return key: string
        """
        return self._key

    def get_parameter(self, parameter_key):
        """Return a parameter based on key.

        :param parameter_key:
        :return: parameter matching the parameter key
        """
        return self._current_parameters[parameter_key]

    def _parameters_up_to_date(self, parameters_store,obfuscator):
        """ Internal method that checks generator and non-generator parameters for being up-to-date.

        :param parameters_store:
        :return: True if no update is needed, otherwise False
        """
        needs_update = not self._store_matches_current(parameters_store,obfuscator)
        if not needs_update:
            for parameter in self._current_parameters.values():
                if isinstance(parameter, parsimony.generators.Generator):
                    needs_update |= not parameter.up_to_date()

        return not needs_update

    def _update_parameters(self):
        """Update generator parameters.
        """
        for parameter_key, parameter in self._current_parameters.items():
            if isinstance(parameter, parsimony.generators.Generator):
                self._current_parameters[parameter_key] = parameter.generate()
